from abc import abstractmethod

import numpy as np

from . import moduleFrame


class Proportionality(moduleFrame.Strategy):
    requiredAttributes = ()

    @abstractmethod
    def run(self, signalVars, totalConcs):
        pass


class GetConcs(Proportionality):
    def run(self, signalVars, totalConcs):
        return signalVars


class GetFraction(Proportionality):
    def run(self, signalVars, totalConcs):
        # divide by total concs
        # TODO: support different signals caused by different parent species
        return signalVars / np.sum(signalVars, axis=1, keepdims=True)


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Experiment"
    dropdownLabelText = "What are the signals proportional to?"
    dropdownOptions = {
        "Concentration (slow exchange)": GetConcs,
        "Mole fraction (fast exchange)": GetFraction,
    }
    attributeName = "proportionality"
    setDefault = False



import tkinter.ttk as ttk
from abc import abstractmethod

import numpy as np
from numpy import ma
from numpy.linalg import lstsq
from scipy.optimize import lsq_linear

from . import moduleFrame
from .table import ButtonFrame, Table


class FitSignals(moduleFrame.Strategy):
    requiredAttributes = ()

    @abstractmethod
    def run(self, signalVars, knownSpectra):
        pass


class FitSignalsOrdinary(FitSignals):
    def run(self, signalVars, knownSpectra):
        # rows are additions, columns are contributors
        knownMask = ~ma.getmaskarray(knownSpectra)[:, 0]
        knownSignals = signalVars[:, knownMask]
        unknownSignals = signalVars[:, ~knownMask]

        # TODO: investigate why this doesn't work with @
        knownSpectrum = np.dot(knownSignals, knownSpectra[knownMask, :])
        unknownSpectrum = self.titration.processedData - knownSpectrum

        if ma.is_masked(unknownSpectrum):
            signalsCount = unknownSpectrum.shape[1]
            unknownsCount = unknownSignals.shape[1]
            fittedSignals = np.empty((unknownsCount, signalsCount))
            residuals = np.empty(signalsCount)

            for index, signal in enumerate(unknownSpectrum.T):
                signalMask = ma.getmaskarray(signal)
                try:
                    fittedSignals[:, index], residuals[index], _, _ = lstsq(
                        unknownSignals[~signalMask, :], signal.compressed(), rcond=None
                    )
                except ValueError:
                    # Should only happen with incorrect constraints
                    fittedSignals[:, index], _, _, _ = lstsq(
                        unknownSignals[~signalMask, :], signal.compressed(), rcond=None
                    )
                    residuals[index] = np.linalg.norm(
                        unknownSignals[~signalMask, :] @ fittedSignals[:, index]
                        - signal.compressed(),
                        ord=2,
                    )
        else:
            fittedSignals, residuals, _, _ = lstsq(
                unknownSignals, np.asarray(unknownSpectrum), rcond=None
            )

        fittedCurves = unknownSignals @ fittedSignals + knownSpectrum
        allSignals = knownSpectra.copy()
        allSignals[~knownMask, :] = fittedSignals
        return allSignals, residuals, fittedCurves


class SignalConstraintsTable(Table):
    def __init__(self, master, titration):
        if hasattr(titration, "signalConstraints"):
            signalConstraints = np.where(
                np.isinf(titration.fitSignals.signalConstraints),
                "",
                titration.fitSignals.signalConstraints,
            )
        else:
            signalConstraints = None

        super().__init__(
            master,
            0,
            0,
            (
                "Lower",
                "Upper",
            ),
            maskBlanks=True,
            rowOptions=(),
            columnOptions="readonlyTitles",
            boldTitles=True,
        )
        self.addRow(data=signalConstraints)


class SignalConstraintsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter signal constraints")

        self.frame = ttk.Frame(self)
        self.frame.pack(expand=True, fill="both")

        constraintsLabel = ttk.Label(
            self.frame, text="Leave cells blank for no lower/upper bound."
        )
        constraintsLabel.pack()

        self.constraintsTable = SignalConstraintsTable(self.frame, titration)
        self.constraintsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.constraintsTable.data = np.array([["", ""]])

    def saveData(self):
        constraints = self.constraintsTable.data[0]

        self.signalConstraints = np.where(
            np.isnan(constraints), (-np.inf, np.inf), constraints
        )

        self.saved = True
        self.destroy()


class FitSignalsConstrained(FitSignals):
    requiredAttributes = FitSignals.requiredAttributes + ("signalConstraints",)

    def run(self, signalVars, knownSpectra):
        # rows are additions, columns are contributors
        knownMask = ~ma.getmaskarray(knownSpectra)[:, 0]
        knownSignals = signalVars[:, knownMask]
        unknownSignals = signalVars[:, ~knownMask]

        knownSpectrum = np.dot(knownSignals, knownSpectra[knownMask, :])
        unknownSpectrum = self.titration.processedData - knownSpectrum

        signalsCount = unknownSpectrum.shape[1]
        unknownsCount = unknownSignals.shape[1]
        fittedSignals = np.empty((unknownsCount, signalsCount))
        residuals = np.empty(signalsCount)

        if ma.is_masked(unknownSpectrum):
            for index, signal in enumerate(unknownSpectrum.T):
                result = lsq_linear(
                    unknownSignals[~ma.getmaskarray(signal), :],
                    signal.compressed(),
                    self.signalConstraints,
                    method="bvls",
                )
                fittedSignals[:, index] = result.x
                residuals[index] = result.cost
        else:
            for index, signal in enumerate(unknownSpectrum.T):
                result = lsq_linear(
                    unknownSignals,
                    signal,
                    self.signalConstraints,
                    method="bvls",
                )
                fittedSignals[:, index] = result.x
                residuals[index] = result.cost

        fittedCurves = unknownSignals @ fittedSignals + knownSpectrum
        allSignals = knownSpectra.copy()
        allSignals[~knownMask, :] = fittedSignals
        return allSignals, residuals, fittedCurves


class FitSignalsNonnegative(FitSignalsConstrained):
    signalConstraints = np.array([0, np.inf])


class FitSignalsCustom(FitSignalsConstrained):
    Popup = SignalConstraintsPopup
    popupAttributes = ("signalConstraints",)


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Signals"
    dropdownLabelText = "Apply constraints to fitted signals?"
    dropdownOptions = {
        "No": FitSignalsOrdinary,
        "Nonnegative": FitSignalsNonnegative,
        "Custom constraints": FitSignalsCustom,
    }
    attributeName = "fitSignals"



import ctypes
import importlib.resources as res
import tkinter as tk
import tkinter.filedialog as fd
import tkinter.messagebox as mb
import tkinter.ttk as ttk
from pathlib import PurePath
from tkinter import font

import numpy as np
import ttkbootstrap
from ttkbootstrap.widgets import InteractiveNotebook

from . import patchMatplotlib, titrationReader, windowsHighDpiPatch
from .style import defaultFigureParams, figureParams, padding
from .table import ButtonFrame
from .titration import Titration
from .titrationFrame import TitrationFrame

patchMatplotlib.applyPatch()
try:
    appId = "daniilS.musketeer"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appId)
except Exception:
    pass


DEBUG = True
olderror = mb.showerror


def newshowerror(title, message, *args, **kwargs):
    import traceback

    if DEBUG and isinstance(message, Exception):
        message = traceback.format_exc()
    return olderror(title, message, *args, **kwargs)


mb.showerror = newshowerror


# need to keep a reference to the Style object so that image-based widgets
# appear correctly
ttkStyle = ttkbootstrap.Style(theme="lumen")
root = ttkStyle.master
root.geometry("1000x600")

windowsHighDpiPatch.setEnhancedDpiScaling(root)

try:
    # Windows and OS X
    root.state("zoomed")
except tk.TclError:
    # X11
    root.attributes("-zoomed", True)
root.title("Musketeer")

try:
    iconData = res.read_binary(__package__, "logo small.png")
    icon = tk.PhotoImage(data=iconData)
    root.iconphoto(True, icon)
except Exception:
    # Should currently never happen, but if anything changes about the API in
    # the future, we can survive without an icon.
    pass
frame = ttk.Frame(root, padding=padding)
frame.pack(expand=True, fill="both")


class DpiPopup(tk.Toplevel):
    def __init__(self, master, saveCallback, *args, **kwargs):
        super().__init__(
            master,
            padx=padding,
            pady=padding,
            *args,
            **kwargs,
        )
        self.title("Change figure DPI")
        self.resizable(False, False)
        self.saveCallback = saveCallback

        scaleLabel = ttk.Label(self, text="Scale:", justify="left", state="normal")
        scaleLabel.grid(row=0, column=0, sticky="ew")

        self.scaleDropdown = ttk.Combobox(
            self,
            values=[f"{i}%" for i in [60, 80, 100, 125, 150, 200]],
            justify="right",
            width=5,
        )
        self.scaleDropdown.grid(row=0, column=2, sticky="ew")

        resLabel = ttk.Label(self, text="Resolution:", justify="left")
        resLabel.grid(row=1, column=0, sticky="ew")

        self.resDropdown = ttk.Combobox(
            self,
            values=[
                f"{int(y*4/3)} x {y}" for y in [240, 360, 480, 600, 960, 1200, 1440]
            ],
            justify="right",
            width=10,
            state="readonly",
        )
        self.resDropdown.grid(row=1, column=2, sticky="ew")

        self.rowconfigure(0, pad=10)
        self.rowconfigure(1, pad=10)

        self.columnconfigure(0, pad=10)
        self.columnconfigure(1, weight=1)

        self.initValues()

        buttonFrame = ButtonFrame(self, self.reset, self.save, self.destroy)
        buttonFrame.saveButton.configure(text="Save")

        buttonFrame.applyButton = ttk.Button(
            buttonFrame, text="Apply", command=self.apply, style="success.TButton"
        )
        buttonFrame.applyButton.pack(
            side="right", after=buttonFrame.saveButton, padx=padding
        )

        buttonFrame.grid(row=2, column=0, columnspan=3, sticky="esw")

    def reset(self):
        self.scaleDropdown.set(f"{defaultFigureParams['scale']}%")
        self.resDropdown.set(f"{defaultFigureParams['x']} x {defaultFigureParams['y']}")

    def initValues(self):
        self.scaleDropdown.set(f"{figureParams['scale']}%")
        self.resDropdown.set(f"{figureParams['x']} x {figureParams['y']}")

    def save(self):
        if self.apply():  # returns False if failed to apply
            self.destroy()

    def apply(self):
        try:
            scale = int(self.scaleDropdown.get().strip("%"))
            x, y = [int(i) for i in self.resDropdown.get().split(" x ")]
            figureParams.update(scale=scale, x=x, y=y)
            self.saveCallback()
            return True
        except Exception as e:
            mb.showerror(title="Could not apply settings", message=e, parent=self)
            return False


class TitrationsNotebook(InteractiveNotebook):
    def __init__(self, master, *args, **kwargs):
        super().__init__(
            master,
            padding=padding,
            newtab=self.newFile,
            style="Flat.Interactive.TNotebook",
            *args,
            **kwargs,
        )
        self.createMenuBar()
        self.createQuickStartMenu()
        self.select(str(self._newtab_frame))

    def forget(self, index, *args, **kwargs):
        # Manually change to a different tab before a tab is closed, to prevent the
        # quickstart menu from appearing briefly right after a tab is closed but before
        # the next tab has been selected
        next_tab = index + 1
        if (
            self._has_newtab_button and next_tab == self._last_tab
        ) or index == self._last_tab:
            next_tab = index - 1
        if 0 <= next_tab <= self._last_tab:
            self.select(next_tab)
        return super().forget(index, *args, **kwargs)

    def createQuickStartMenu(self):
        self.quickStartFrame = ttk.Frame(self)
        self.quickStartFrame.pack(expand=True, fill="none")

        self.quickStartFont = font.nametofont("TkTextFont").copy()
        self.quickStartFont["size"] = round(self.quickStartFont["size"] * 1.33)
        ttkStyle.configure("large.success.TButton", font=self.quickStartFont)
        ttkStyle.configure("large.TButton", font=self.quickStartFont)

        self.newFileButton = ttk.Button(
            self.quickStartFrame,
            text="Create a new fit file",
            command=self.newFile,
            style="large.success.TButton",
        )
        self.newFileButton.pack(fill="x")
        self.orLabel = ttk.Label(
            self.quickStartFrame, text="or", font=self.quickStartFont
        )
        self.orLabel.pack(pady=5)
        self.OpenFileButton = ttk.Button(
            self.quickStartFrame,
            text="Open an existing file",
            command=self.openFile,
            style="large.TButton",
        )
        self.OpenFileButton.pack(fill="x")

    def createMenuBar(self):
        self.menuBar = tk.Menu(self, tearoff=False)

        fileMenu = tk.Menu(self.menuBar, tearoff=False)
        self.menuBar.add_cascade(label="File", menu=fileMenu, underline=0)

        self.addMenuCommand(fileMenu, "New", self.newFile)
        self.addMenuCommand(fileMenu, "Open", self.openFile)
        self.addMenuCommand(fileMenu, "Save", self.saveFile)
        self.addMenuCommand(
            fileMenu, "Save As", self.saveFileAs, keys=("Shift", "s"), underline=5
        )

        editMenu = tk.Menu(self.menuBar, tearoff=False)
        self.menuBar.add_cascade(label="Edit", menu=editMenu, underline=0)
        editMenu.add_command(
            label="Change figure DPI", command=self.editDpi, underline=0
        )

        self.winfo_toplevel().config(menu=self.menuBar)

    def addMenuCommand(self, fileMenu, label, command, keys=None, underline=0):
        if self._windowingsystem == "aqua":
            accelerator = "Command"
            key = accelerator
        else:
            accelerator = "Ctrl"
            key = "Control"

        if keys is None:
            keys = (label[0].lower(),)

        fileMenu.add_command(
            label=label,
            command=command,
            accelerator=f"{accelerator}+{'+'.join(keys).title()}",
            underline=underline,
        )
        self.winfo_toplevel().bind(f"<{key}-{'-'.join(keys)}>", command)

    def newFile(self, *args):
        titration = Titration("Fit 1")
        titration.rawData = np.empty((0, 0))

        titrationFrame = TitrationFrame(self, padding=padding)
        self.add(titrationFrame, text="New Titration", sticky="nesw")
        self.select(str(titrationFrame))
        titrationFrame.loadTitration(titration)

    def saveFile(self, *args):
        self.nametowidget(self.select()).saveFile()

    def saveFileAs(self, *args):
        self.nametowidget(self.select()).saveFile(saveAs=True)

    def openFile(self, *args):
        fileType = tk.StringVar(self)
        fileReaders = titrationReader.fileReaders
        filePath = fd.askopenfilename(
            title="Open file",
            filetypes=fileReaders[:, :-1].tolist(),
            typevariable=fileType,
        )
        if filePath == "":
            return
        fileReader = fileReaders[fileReaders[:, 0] == fileType.get(), -1].item()
        try:
            titrations = fileReader(filePath)
        except Exception as e:
            mb.showerror(title="Failed to read file", message=e, parent=self)
            return

        self.tk.eval("tk busy .")
        self.update()
        # create a tab for each titration, and let the titration object handle
        # its own I/O
        for titration in titrations:
            titrationFrame = TitrationFrame(self, filePath, padding=padding)
            self.add(titrationFrame, text=PurePath(filePath).name, sticky="nesw")
            self.select(str(titrationFrame))
            titrationFrame.loadTitration(titration)
        self.tk.eval("tk busy forget .")

    def editDpi(self, *args):
        popup = DpiPopup(self, self.updateDpi)
        popup.withdraw()
        self.update()
        root = self.winfo_toplevel()
        x = root.winfo_x() + root.winfo_width() / 2 - popup.winfo_width() / 2
        y = root.winfo_y() + root.winfo_height() / 2 - popup.winfo_height() / 2
        popup.geometry(f"+{int(x)}+{int(y)}")
        popup.transient(self)
        popup.grab_set()
        popup.deiconify()
        popup.wait_window()

    def updateDpi(self):
        for tab in self.tabs():
            if isinstance(titrationFrame := self.nametowidget(tab), TitrationFrame):
                titrationFrame.updateDpi()
        self.update()


notebook = TitrationsNotebook(frame)
notebook.pack(expand=True, fill="both")

root.mainloop()



import tkinter as tk
import tkinter.ttk as ttk
from abc import abstractmethod

import numpy as np
from numpy import ma

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .style import padding
from .table import ButtonFrame, Table, WrappedLabel


class Contributors(moduleFrame.Strategy):
    requiredAttributes = (
        "contributorsMatrix",
        "contributorNames",
    )

    @abstractmethod
    def run(self, freeConcs, boundConcs):
        pass


class GetSignalVarsFromMatrix(Contributors):
    def run(self, freeConcs, boundConcs):
        allConcs = np.concatenate((freeConcs, boundConcs), 1)
        variableConcs = allConcs @ self.contributorsMatrix.T
        return variableConcs


class ContributorsTable(Table):
    def __init__(self, master, rowTitles, columnTitles, data):
        self.width = max([len(title) for title in columnTitles])
        super().__init__(
            master,
            0,
            0,
            columnTitles,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
        )
        for name, contributions in zip(rowTitles, data):
            self.addRow(name, contributions)

    def newRow(self):
        defaultEntries = np.full(self.dataCells.shape[1], "0")
        self.addRow("New state", defaultEntries)

    def convertData(self, number):
        # if all entries happen to be integers, this is handled by saveData()
        return float(number)


class ContributorsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing states")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        contributorsLabel = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "On each row, enter a state that contributes to the observed"
                " signal. For each column, specify how many of the state that"
                " species contains."
            ),
        )
        contributorsLabel.pack(expand=False, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        scrolledFrame = ScrolledFrame(self.frame, max_width=1500)
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)
        self.createTable()

    def createTable(self):
        titration = self.titration
        if (
            titration.contributors.contributorsMatrix.shape[1]
            == titration.speciation.outputCount
        ):
            contributorsMatrix = titration.contributors.contributorsMatrix
        else:
            contributorsMatrix = np.eye(titration.speciation.outputCount)

        self.contributorsTable = ContributorsTable(
            self.innerFrame,
            titration.contributors.contributorNames,
            titration.speciation.outputNames,
            contributorsMatrix,
        )
        self.contributorsTable.pack(expand=True, fill="both")

    def reset(self):
        self.contributorsTable.destroy()
        self.createTable()

    def saveData(self):
        self.contributorsMatrix = self.contributorsTable.data
        if np.all(self.contributorsMatrix % 1 == 0):
            self.contributorsMatrix = self.contributorsMatrix.astype(int)

        self.contributorNames = self.contributorsTable.rowTitles

        self.saved = True
        self.destroy()


class GetSignalVarsCustom(GetSignalVarsFromMatrix):
    Popup = ContributorsPopup
    popupAttributes = ("contributorsMatrix", "contributorNames")


class ContributorsPerSpeciesNotebook(ttk.Notebook):
    def __init__(self, master, titration, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.titration = titration

        mapTab = ttk.Frame(self)
        self.add(mapTab, text="Map signal to species", sticky="nsew", padding=padding)
        contributorsLabel = WrappedLabel(
            mapTab,
            padding=padding * 2,
            text="For each signal, specify which species it is caused by.",
        )
        contributorsLabel.pack(expand=False, fill="both")

        if (
            hasattr(titration.contributors, "contributorsMatrixPerSpecies")
            and titration.contributors.contributorsMatrixPerSpecies.shape[2]
            == titration.speciation.outputCount
        ):
            contributorsMatrixPerSpecies = (
                titration.contributors.contributorsMatrixPerSpecies
            )
            contributorNamesPerSpecies = (
                titration.contributors.contributorNamesPerSpecies
            )
        else:
            (
                contributorsMatrixPerSpecies,
                contributorNamesPerSpecies,
            ) = self.createDefaultSetup()

        if (
            hasattr(titration.contributors, "signalToSpeciesMap")
            and len(titration.contributors.signalToSpeciesMap) == titration.numSignals
            and np.all(
                titration.contributors.signalToSpeciesMap
                <= titration.totalConcentrations.freeCount
            )
        ):
            signalToSpeciesMap = titration.contributors.signalToSpeciesMap
        else:
            signalToSpeciesMap = np.zeros(titration.numSignals, dtype=int)

        self.speciesTabs = []
        self.mapVars = []

        radioFrame = ttk.Frame(mapTab)
        radioFrame.pack(expand=True, fill="both")

        for i, (freeName, contributorNames, contributorsMatrix) in enumerate(
            zip(
                titration.totalConcentrations.freeNames,
                contributorNamesPerSpecies,
                contributorsMatrixPerSpecies,
            )
        ):
            label = ttk.Label(radioFrame, text=freeName)
            label.grid(row=0, column=i + 1, sticky="w")
            radioFrame.columnconfigure(i + 1, uniform="map", pad=padding)

            rowTitles = contributorNames[~contributorNames.mask]
            columnTitles = titration.speciation.outputNames[self.filter(i).astype(bool)]
            matrix = contributorsMatrix[~contributorsMatrix.mask].reshape(
                (len(rowTitles), len(columnTitles))
            )
            tab = ttk.Frame(self)
            contributorsLabel = WrappedLabel(
                tab,
                padding=padding * 2,
                text=(
                    "On each row, enter a state that contributes to the observed"
                    " signal. For each column, specify how many of the state that"
                    " species contains."
                ),
            )
            contributorsLabel.pack(expand=False, fill="both")

            scrolledFrame = ScrolledFrame(tab, max_width=1500)
            scrolledFrame.pack(expand=True, fill="both")
            innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)

            table = ContributorsTable(innerFrame, rowTitles, columnTitles, matrix)
            table.pack(expand=True, fill="both")
            self.add(tab, text=freeName, state="hidden", sticky="nsew", padding=padding)
            self.speciesTabs.append(tab)

        for i, (title, index) in enumerate(
            zip(titration.processedSignalTitlesStrings, signalToSpeciesMap)
        ):
            label = ttk.Label(radioFrame, text=title)
            label.grid(row=i + 1, column=0, padx=padding)

            mapVar = tk.StringVar(self)
            self.mapVars.append(mapVar)
            for j in range(len(titration.totalConcentrations.freeNames)):
                radioButton = ttk.Radiobutton(
                    radioFrame, variable=mapVar, value=j, command=self.updateTabs
                )
                radioButton.grid(row=i + 1, column=j + 1, sticky="w")
                if j == index:
                    radioButton.invoke()

    def updateTabs(self):
        # show tabs that have any signals mapped to them, hide the rest
        for i, tab in enumerate(self.speciesTabs):
            if any(mapVar.get() == str(i) for mapVar in self.mapVars):
                self.tab(tab, state="normal")
            else:
                self.tab(tab, state="hidden")

    def createDefaultSetup(self):
        contributorsMatrixPerSpecies = ma.array(
            np.empty(
                (
                    self.titration.totalConcentrations.freeCount,
                    self.titration.totalConcentrations.freeCount,
                    self.titration.totalConcentrations.outputCount,
                ),
                dtype=int,
            ),
            mask=True,
        )
        contributorNamesPerSpecies = ma.array(
            np.empty(
                (
                    self.titration.totalConcentrations.freeCount,
                    self.titration.totalConcentrations.freeCount,
                ),
                dtype=object,
            ),
            mask=True,
        )

        for i, name in enumerate(self.titration.totalConcentrations.freeNames):
            filter = self.filter(i)
            contributorsMatrixPerSpecies[i, i, filter.astype(bool)] = 1
            contributorNamesPerSpecies[i, i] = name

        return contributorsMatrixPerSpecies, contributorNamesPerSpecies

    def filter(self, index):
        free = np.zeros(self.titration.totalConcentrations.freeCount, dtype=int)
        free[index] = 1
        bound = abs(self.titration.speciation.stoichiometries[:, index])
        return np.concatenate((free, bound))


class ContributorsPerSpeciesPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing species")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")
        self.notebook = ContributorsPerSpeciesNotebook(
            self.frame, titration, style="Flat.TNotebook"
        )
        self.notebook.pack(expand=True, fill="both", padx=padding, pady=padding)

    def reset(self):
        self.notebook.destroy()
        self.notebook = ContributorsPerSpeciesNotebook(self.frame, self.titration)
        self.notebook.pack(expand=True, fill="both")

    def saveData(self):
        raise NotImplementedError

        self.saved = True
        self.destroy()


class GetSignalVarsPerSpecies(Contributors):
    Popup = ContributorsPerSpeciesPopup
    popupAttributes = (
        "contributorsMatrixPerSpecies",
        "contributorNamesPerSpecies",
        "signalToSpeciesMap",
    )

    contributorsMatrix = None
    contributorNames = None

    def run(*args):
        # TODO: implement
        raise NotImplementedError


class GetSignalVarsAll(Contributors):
    @property
    def contributorNames(self):
        return self.titration.speciation.outputNames

    @property
    def contributorsMatrix(self):
        return np.identity(self.titration.speciation.outputCount, dtype=int)

    def run(self, freeConcs, boundConcs):
        allConcs = np.concatenate((freeConcs, boundConcs), 1)
        return allConcs


class GetSignalVarsSingle(GetSignalVarsFromMatrix):
    requiredAttributes = GetSignalVarsFromMatrix.requiredAttributes + (
        "contributorIndex",
    )

    # TODO: make this work with polymers
    def filter(self):
        hostFree = np.zeros(self.titration.speciation.freeCount, dtype=int)
        hostFree[self.contributorIndex] = 1

        hostBound = abs(
            self.titration.speciation.stoichiometries[:, self.contributorIndex]
        )

        return np.concatenate((hostFree, hostBound))

    @property
    def contributorNames(self):
        return self.titration.speciation.outputNames[self.filter().astype(bool)]

    @property
    def contributorsMatrix(self):
        matrix = np.diag(self.filter())
        emptyRows = np.all(matrix == 0, axis=1)
        return matrix[~emptyRows]


class GetSignalVarsHost(GetSignalVarsSingle):
    contributorIndex = 0


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Signals"
    dropdownLabelText = "What contributes to the signals?"
    dropdownOptions = {
        "Only Host": GetSignalVarsHost,
        "All": GetSignalVarsAll,
        "Custom": GetSignalVarsCustom,
        # "Custom per species": GetSignalVarsPerSpecies,
    }
    attributeName = "contributors"
